#' Local linear regression of a scalar response against a matrix of covariates 
#'
#'
#' @param Responses A numeric vector of length n containing the responses Y1,...,Yn.
#' @param COVARIATES A numeric matrix of n rows and p columns containing the predictors.
#' @param Bwknn.ratio A numeric in the (0,1] range containing the $k$-nearest-neighbour (KNN) ratio bandwidth to be used.
#' @param standard A boolean. If \code{standard = TRUE} (the default), the matrix \code{COVARIATES} is standardised.
#' @return An llreg S3 class. 
llreg = function(Responses, COVARIATES, Bwknn.ratio, standard = T)
{
  
  COVARIATES = as.matrix(COVARIATES)
  if(!is.matrix(COVARIATES))
  {
    stop("cannot convert COVARIATES to a matrix")
  }
  if(!all(is_not_na(COVARIATES)))
  {
    stop("COVARIATES contains NA values")
  }
  if(!all(is_not_null(COVARIATES)))
  {
    stop("COVARIATES contains NULL values")
  }
  if(!is_numeric(COVARIATES))
  {
    stop("COVARIATES must be of type numeric")
  }
  if(Reduce("|",COVARIATES == Inf))
  {
    stop("COVARIATES contains Inf values")
  }
  
  Responses = matrix(as.vector(Responses),ncol=1)
  
  if(!is.matrix(Responses))
  {
    stop("cannot convert Responses to a n by 1 matrix")
  }
  if(!all(is_not_na(Responses)))
  {
    stop("Responses contains NA values")
  }
  if(!all(is_not_null(Responses)))
  {
    stop("Responses contains NULL values")
  }
  if(!is_numeric(Responses))
  {
    stop("Responses must be of type numeric")
  }
  if(Reduce("|",Responses == Inf))
  {
    stop("Responses contains Inf values")
  }
  
  if(dim(COVARIATES)[1] != dim(Responses)[1]){
    stop("COVARIATES and Responses must have the same number of rows.")
  }
  
  if(is.null(Bwknn.ratio)){
    stop("Bwknn.ratio is a NULL object")
  }
  
  if (length(Bwknn.ratio) != 1){
    stop("Bwknn.ratio must be of length 1.")
  }
  
  if(!all(is_not_na(Bwknn.ratio)))
  {
    stop("Bwknn.ratio contains NA values")
  }
  if(!all(is_not_null(Bwknn.ratio)))
  {
    stop("Bwknn.ratio contains NULL values")
  }
  if(!is_numeric(Bwknn.ratio))
  {
    stop("Bwknn.ratio must be of type numeric")
  }
  if(Reduce("|",Bwknn.ratio == Inf))
  {
    stop("Bwknn.ratio contains Inf values")
  }
  
  if(sum(Bwknn.ratio <= 0) + sum(Bwknn.ratio > 1) > 0){
    stop("Bwknn.ratio must be in the (0,1] interval")
  }
  
  if(is.null(standard)){
    stop("standard is a NULL object")
  }
  
  if (length(standard) != 1){
    stop("standard must be of length 1.")
  }
  
  if(!all(is_not_na(standard)))
  {
    stop("standard contains NA values")
  }
  if(!is_logical(standard))
  {
    stop("standard must be of type logical")
  }
  
  
  if(standard){
    Meancovariates = apply(COVARIATES, 2, mean)
    CCOVARIATES = t(t(COVARIATES) - Meancovariates)
    Stdev = sqrt(apply(CCOVARIATES^2, 2, mean))
    STDCOVAR = t(t(CCOVARIATES) / Stdev)
  }else{
    STDCOVAR = COVARIATES
  }
  Responses = matrix(as.vector(Responses),ncol = 1)
  Output    = llreg2(Responses, STDCOVAR, Bwknn.ratio)
  names(Output) = c("Interpolated.values" , "loocv", "bwknn.ratio")
  Output[["Responses"]]  = Responses
  Output[["COVARIATES"]] = COVARIATES
  Output[["standard"]]   = standard
  return(structure(Output,class="llreg"))
}