#' Most-predictive covariate-selection for nonparametric regression of a scalar response against multivariate predictors.
#'
#'
#' NOnparametric VAriable Selection (NOVAS) by Ferraty and Hall (2015) selects the most-predictive covariates by means of a stepwise forward strategy: At each step the most-predictive covariate 
# is selected and dropped from the set of covariates in the next step. 
#' 
#' @param Responses A numeric vector of length n containing the responses Y1,...,Yn.
#' @param COVARIATES A numeric matrix of n rows and p columns containing the predictors.
#' @param Bwknn.ratio A vector of numerics in the [0,1] range containing the $k$-nearest-neighbour (KNN) ratio bandwidths to be investigated. 
#' It defaults to \code{Bwknn.ratio = c(0.1,0.2,0.3,0.4,0.5,0.6,0.7)}.
#' @param maxsteps A positive integer giving a bound on the cardinality of the sets of covariates to be investigated. It defaults to 10.
#' @param exponent A numeric in the [0,1] range governing the efficiency vs computational complexity tradeoff. At each iteration $p^exponent$ models will be maintained; lower values of \code{exponent} therefore make Novas faster. The default value, 0.5, corresponds to the choice of parameter made in Ferraty and Hall (2015).
#' @param threshold A numeric in the [0,1] range giving the stopping criterion. NOVAS will stop when the improvement 
#' gained from adding an additional covariate is smaller than (100 * threshold) \%. The default is 0.05.
#' @param standard A boolean. If \code{standard = TRUE} (the default), the matrix \code{COVARIATES} is standardised.
#' @param ploocvpar A non-negative numeric indicating the penalty to be added to the leave one out (LOO) CV. When set to zero (the default) un-penalised LOO-CV is used.
#' @param screening A boolean. If \code{screening = TRUE} the selected set of covariates along with the evaluation criterion are printed after each iteration.
#' 
#' @examples 
#' ### Simulated data example
#' 
#' library(Novas)
#' set.seed(1)
#' 
#' n = 200
#' p = 3
#' COVARIATES = matrix(rnorm(n*p),nrow = n,ncol = p)
#' Responses  = matrix(COVARIATES[,2]^2,ncol=1)+rnorm(n)
#' Out = novas(Responses, COVARIATES,ploocvpar = 0.05)
#' Robust_Out = robustify(Out)
#' summary(Robust_Out)
#' 
#' ### Real data example
#' 
#' library(Novas)
#' data(pork)
#' 
#' pork_response       = pork$pork_response
#' pork_spec_curves    = pork$pork_spec_curves
#' 
#' pork_results = novas(pork_response, pork_spec_curves,threshold = 0.2)
#' 
#' # visualise the most predictive wavelengths
#' par(mfrow = c(1, 1), mar = c(5, 5, 1, 1), cex.axis = 1.5, cex.lab = 1.5)
#' spec_curves.ts = ts(t(pork_spec_curves), start = 850, frequency = 0.5)
#' ts.plot(spec_curves.ts, gpars = list(col = gray(0.5), xlab = "wavelengths", ylab = "absorbance"))
#' x = seq(850, by = 2, length = 100)
#' abline(v = x[pork_results$model], lty = 1, lwd = 2)
#' 
#' 
#' @references \insertRef{ferraty2015algorithm}{Novas}
#' 
novas = function(Responses, COVARIATES, Bwknn.ratio = c(0.1,0.2,0.3,0.4,0.5,0.6,0.7), maxsteps = 10, exponent = 0.5, threshold = 0.05, standard = T, ploocvpar = 0,
                 screening = T)
{
  
  COVARIATES = as.matrix(COVARIATES)
  if(!is.matrix(COVARIATES))
  {
    stop("cannot convert COVARIATES to a matrix")
  }
  if(!all(is_not_na(COVARIATES)))
  {
    stop("COVARIATES contains NA values")
  }
  if(!all(is_not_null(COVARIATES)))
  {
    stop("COVARIATES contains NULL values")
  }
  if(!is_numeric(COVARIATES))
  {
    stop("COVARIATES must be of type numeric")
  }
  if(Reduce("|",COVARIATES == Inf))
  {
    stop("COVARIATES contains Inf values")
  }
  
  Responses = matrix(as.vector(Responses),ncol=1)
  
  if(!is.matrix(Responses))
  {
    stop("cannot convert Responses to a n by 1 matrix")
  }
  if(!all(is_not_na(Responses)))
  {
    stop("Responses contains NA values")
  }
  if(!all(is_not_null(Responses)))
  {
    stop("Responses contains NULL values")
  }
  if(!is_numeric(Responses))
  {
    stop("Responses must be of type numeric")
  }
  if(Reduce("|",Responses == Inf))
  {
    stop("Responses contains Inf values")
  }
  
  if(dim(COVARIATES)[1] != dim(Responses)[1]){
    stop("COVARIATES and Responses must have the same number of rows.")
  }
  
  if(is.null(Bwknn.ratio)){
    stop("Bwknn.ratio is a NULL object")
  }
  
  if(!all(is_not_na(Bwknn.ratio)))
  {
    stop("Bwknn.ratio contains NA values")
  }
  if(!all(is_not_null(Bwknn.ratio)))
  {
    stop("Bwknn.ratio contains NULL values")
  }
  if(!is_numeric(Bwknn.ratio))
  {
    stop("Bwknn.ratio must be of type numeric")
  }
  if(Reduce("|",Bwknn.ratio == Inf))
  {
    stop("Bwknn.ratio contains Inf values")
  }
  
  if(sum(Bwknn.ratio <= 0) + sum(Bwknn.ratio > 1) > 0){
    stop("Bwknn.ratio must be in the (0,1] interval")
  }
  
  if(is.null(standard)){
    stop("standard is a NULL object")
  }
  
  if (length(standard) != 1){
    stop("standard must be of length 1.")
  }
  
  if(!all(is_not_na(standard)))
  {
    stop("standard contains NA values")
  }
  if(!is_logical(standard))
  {
    stop("standard must be of type logical")
  }
  
  if(is.null(screening)){
    stop("screening is a NULL object")
  }
  
  if (length(screening) != 1){
    stop("screening must be of length 1.")
  }
  
  if(!all(is_not_na(screening)))
  {
    stop("screening contains NA values")
  }
  if(!is_logical(screening))
  {
    stop("screening must be of type logical")
  }
  
  if(is.null(threshold)){
    stop("threshold is a NULL object")
  }
  
  if (length(threshold) != 1){
    stop("threshold must be of length 1.")
  }
  
  if(!all(is_not_na(threshold)))
  {
    stop("threshold contains NA values")
  }
  if(!all(is_not_null(threshold)))
  {
    stop("threshold contains NULL values")
  }
  if(!is_numeric(threshold))
  {
    stop("threshold must be of type numeric")
  }
  if(Reduce("|",threshold == Inf))
  {
    stop("threshold contains Inf values")
  }
  
  if(sum(threshold <= 0) + sum(threshold > 1) > 0){
    stop("threshold must be in the (0,1] interval")
  }
  
  if(is.null(ploocvpar)){
    stop("ploocvpar is a NULL object")
  }
  
  if (length(ploocvpar) != 1){
    stop("ploocvpar must be of length 1.")
  }
  
  if(!all(is_not_na(ploocvpar)))
  {
    stop("ploocvpar contains NA values")
  }
  if(!all(is_not_null(ploocvpar)))
  {
    stop("ploocvpar contains NULL values")
  }
  if(!is_numeric(ploocvpar))
  {
    stop("ploocvpar must be of type numeric")
  }
  if(Reduce("|",ploocvpar == Inf))
  {
    stop("ploocvpar contains Inf values")
  }
  
  if(sum(ploocvpar < 0)){
    stop("ploocvpar must be greater or equal to 0")
  }
  
  if(is.null(exponent)){
    stop("exponent is a NULL object")
  }
  
  if (length(exponent) != 1){
    stop("exponent must be of length 1.")
  }
  
  if(!all(is_not_na(exponent)))
  {
    stop("exponent contains NA values")
  }
  if(!all(is_not_null(exponent)))
  {
    stop("exponent contains NULL values")
  }
  if(!is_numeric(exponent))
  {
    stop("exponent must be of type numeric")
  }
  if(Reduce("|",exponent == Inf))
  {
    stop("exponent contains Inf values")
  }
  
  if(sum(exponent < 0)){
    stop("exponent must be greater or equal to 0")
  }
  
  if(is.null(maxsteps)){
    stop("maxsteps is a NULL object")
  }
  
  if (maxsteps - as.integer(maxsteps) != 0){
    maxsteps = as.integer(maxsteps)
    warning("maxsteps is not an integer and has been rounded down")
  }
  
  if (length(maxsteps) != 1){
    stop("maxsteps must be of length 1.")
  }
  
  if(!all(is_not_na(maxsteps)))
  {
    stop("maxsteps contains NA values")
  }
  if(!all(is_not_null(maxsteps)))
  {
    stop("maxsteps contains NULL values")
  }
  if(!is_numeric(maxsteps))
  {
    stop("maxsteps must be of type numeric")
  }
  if(Reduce("|",maxsteps == Inf))
  {
    stop("maxsteps contains Inf values")
  }
  
  if(sum(maxsteps < 1)){
    stop("maxsteps must be greater or equal to 1")
  }
  
  j<-kk<-NULL

  Responses = matrix(as.vector(Responses),ncol=1)
  COVARIATES = as.matrix(COVARIATES)
  
  if(standard){
    Meancovariates = apply(COVARIATES, 2, mean)
    CCOVARIATES = t(t(COVARIATES) - Meancovariates)
    Stdev = sqrt(apply(CCOVARIATES^2, 2, mean))
    STDCOVAR = t(t(CCOVARIATES) / Stdev)
  }else{
    STDCOVAR = COVARIATES
  }

  p = ncol(COVARIATES)
  n = nrow(COVARIATES)
  
  args = list(Bwknn.ratio = Bwknn.ratio, maxsteps = maxsteps, exponent = exponent, threshold = threshold,screening=screening,ploocvpar=ploocvpar)
  
  if( p > 1){
    Ploocvmin = 0
    retainednb = ceiling(p ** exponent)
    #############################################################################
    # STEP 1: ranking variables according to the CV criterion
    #############################################################################
    Loocv = foreach(j = 1:p, .combine = 'c', .export = "llreg2") %do% {
      loocv = llreg2(Responses, matrix(STDCOVAR[, j],ncol=1), Bwknn.ratio)[[2]]
      return(loocv)
    }
    Ploocv = Loocv * (1 + ploocvpar / log(n))
    Ordered = order(Ploocv)
    Ploocvmin[1] = Ploocv[Ordered[1]]
    oldlistofmodels = as.list(Ordered)
    Oldploocv = Ploocv[Ordered]
    if(screening) cat("Step 1: Selected model =", Ordered[1], " and ploocv =", Oldploocv[1],"\n")
    ##############################################################################
    # STEP >=2: selecting pairs, triples, quadruples, etc  
    ##############################################################################
    iter = 1
    epsilon = 1.5 * threshold
    maxiter = min(maxsteps, p)
    while((epsilon > threshold) && (iter < maxiter)){
      iter = iter + 1
      INDEXES = matrix(rep(1:retainednb, retainednb), nrow = retainednb)
      Index1 = rep(1:(retainednb - 1), (retainednb - 1):1)
      Index2 = as.vector(INDEXES[row(INDEXES) > col(INDEXES)])
      PAIRS = cbind(Index1, Index2)
      newlistofmodels = foreach(kk = 1:nrow(PAIRS)) %do% {
        sort(unique(c(oldlistofmodels[[PAIRS[kk, 1]]], oldlistofmodels[[PAIRS[kk, 2]]])))
      }
      newlistofmodels = unique(newlistofmodels)
      llist = length(newlistofmodels)
      retainednb = min(llist, retainednb)
      if(retainednb >= 1){
        Ploocv = 0
        Ploocv = foreach(kk = 1:llist, .combine='c', .export="llreg2") %do% {
          llistkk=length(newlistofmodels[[kk]])
          Ploocv = llreg2(Responses, STDCOVAR[, newlistofmodels[[kk]]], Bwknn.ratio)[[2]] * (1 + llistkk * ploocvpar / log(n))
          return(Ploocv)
        }
        Ordered = order(Ploocv)
        Ploocvmin[iter] = Ploocv[Ordered[1]]
        epsilon = (Ploocvmin[iter - 1] - Ploocvmin[iter])/Ploocvmin[iter - 1]
        if(epsilon > threshold){
          if(screening) cat("Step", iter, ": Selected model =", newlistofmodels[[Ordered[1]]], " and ploocv = ", Ploocvmin[iter],"\n")
          oldlistofmodels = newlistofmodels[Ordered]
          Oldploocv = Ploocv[Ordered]
        }
      }else{
        epsilon = 0.5 * threshold
      }
    }
    
    Selectedmodel = sort(oldlistofmodels[[1]])
   
    
  } else{
    
    Selectedmodel = 1
    
  }
  
  model.size    = length(Selectedmodel)
  res.llreg     = llreg2(Responses, matrix(STDCOVAR[, Selectedmodel],ncol=model.size), Bwknn.ratio)
  names(res.llreg) = c("Interpolated.values" , "loocv", "bwknn.ratio")
  res.llreg[["Responses"]]  = Responses
  res.llreg[["COVARIATES"]] = matrix(COVARIATES[, Selectedmodel],ncol=model.size)
  res.llreg[["standard"]]   = standard
  
  ploocvmin = res.llreg$loocv*(1 + model.size * ploocvpar / log(n))
  
  if(screening) cat("Final selected model =", Selectedmodel, " and ploocv =", ploocvmin,"\n")
  
  object = list(llreg = structure(res.llreg,class="llreg"), COVARIATES = COVARIATES, model = Selectedmodel, args = args, ploocvmin = ploocvmin)		
  
  class(object) = "novas"
  
  return(structure(object,class="novas"))
  
}